/**
 * @file kernel/pdevs/qss/Derivative.hpp
 * @author The ARTIS Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * ARTIS - the multimodeling and simulation environment
 * This file is a part of the ARTIS environment
 *
 * Copyright (C) 2013-2019 ULCO http://www.univ-littoral.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef QSS_DERIVATIVE
#define QSS_DERIVATIVE

#include <artis-star/kernel/pdevs/Dynamics.hpp>
#include <artis-star/kernel/pdevs/qss/Data.hpp>

namespace artis {
    namespace pdevs {
        namespace qss {

            template<class Time, class Dyn, class Parameters = common::NoParameters>
            class Derivative : public artis::pdevs::Dynamics<Time, Dyn, Parameters> {
                typedef enum {
                    INIT = 0, WAIT, RESPONSE
                } State;

            public:
                enum inputs {
                    IN = 0
                };

                enum outputs {
                    OUT = 0
                };

                enum states {
                    STATE = 0, INPUT_NUMBER, OUTPUT_VALUE, LAST_OUTPUT
                };

                typedef Derivative<Time, Dyn, Parameters> type;

                Derivative(const std::string& name, const Context<Time, Dyn, Parameters>& context)
                        :
                        artis::pdevs::Dynamics<Time, Dyn, Parameters>(name, context),
                        _external_number(0), _internal_number(0)
                {
                    DECLARE_STATES(int,
                            ((STATE, &type::_state)));
                    DECLARE_STATES(unsigned int,
                            ((INPUT_NUMBER, &type::_input_number)));
                    DECLARE_STATES(double,
                            ((OUTPUT_VALUE, &type::_output_value),
                                    (LAST_OUTPUT, &type::_output_value)));

                    this->input_port({IN, "in"});
                    this->output_port({OUT, "out"});
                }

                virtual ~Derivative() { }

                int external(const std::string& name, double Dyn::* var)
                {
                    ++_external_number;
                    this->state_(LAST_OUTPUT + _external_number + 1, name, var);
                    this->input_port({IN + _external_number, name});
                    return IN + _external_number;
                }

                void internal(const std::string& name, double Dyn::* var)
                {
                    assert(_internal_number == 0);

                    ++_internal_number;
                    this->state_(LAST_OUTPUT + 1, name, var);
                }

                virtual double compute() const = 0;

                virtual void dconf(typename Time::type t, typename Time::type e,
                        const common::Bag<Time>& bag)
                {
                    dint(t);
                    dext(t, e, bag);
                }

                virtual void dint(typename Time::type /* time */)
                {
                    if (_state == RESPONSE) {
                        _last_output = _output_value;
                    }
                    _state = WAIT;
                }

                virtual void dext(typename Time::type t, typename Time::type e,
                        const common::Bag<Time>& bag)
                {
                    std::for_each(bag.begin(), bag.end(),
                            [this, t, e](const common::ExternalEvent<Time>& event) {
                                IntegratorData data;

                                event.data()(data);
                                this->get(event.port_index() + LAST_OUTPUT + 1).put(
                                        dynamic_cast<Dyn*>(this),
                                        data.value);
                                switch (_state) {
                                case INIT:
                                    if (_input_number == this->state_number() - (LAST_OUTPUT + 1)) {
                                        _output_value = compute();
                                        _state = RESPONSE;
                                    }
                                    break;
                                case WAIT:
                                case RESPONSE:
                                    double value = compute();

                                    if (value != _last_output) {
                                        _output_value = value;
                                        _state = RESPONSE;
                                    } else {
                                        _state = WAIT;
                                    }
                                }
                            });
                }

                virtual void start(typename Time::type /* time */)
                {
                    _input_number = this->input_port_number();
                    if (_input_number == 0) {
                        _output_value = compute();
                        _state = RESPONSE;
                    } else {
                        _state = INIT;
                    }
                }

                virtual typename Time::type ta(typename Time::type /* time */)
                {
                    switch (_state) {
                    case INIT:
                        return Time::infinity;
                    case WAIT:
                        return Time::infinity;
                    case RESPONSE:
                        return 0;
                    }
                    return Time::infinity;
                }

                virtual common::Bag<Time> lambda(typename Time::type /* time */) const
                {
                    common::Bag<Time> msgs;

                    switch (_state) {
                    case INIT:
                        break;
                    case WAIT:
                        break;
                    case RESPONSE:
                        const DerivativeData data = {_output_value};

                        msgs.push_back(common::ExternalEvent<Time>(OUT, data));
                    }
                    return msgs;
                }

                virtual common::Value observe(const typename Time::type& /* t */,
                        unsigned int /* index */) const
                {
                    return common::Value();
                }

            private:
                unsigned int _external_number;
                unsigned int _internal_number;

                // state
                int _state;
                unsigned int _input_number;
                double _output_value;
                double _last_output;
            };

        }
    }
}

#endif
